"""
A well documented and typed API wrapper for https://hentaihaven.xxx.

Documentation: https://jokelbaf.github.io/hhaven

Source Code: https://github.com/jokelbaf/hhaven
"""

from . import models, exceptions
from .client import *

__version__ = "0.1.4"



"""Main client module."""

import logging, aiohttp, typing, json, aiocache
from collections import namedtuple

from . import exceptions, utility, models
from .decorators import requires_build, requires_token, cached


__all__ = [
    "Client"
]


class Client:
    """Hentai Haven client."""
    
    # Custom docs for pydoc
    token: str
    """Cloudflare token used to access Hentai Haven API."""
    cache: aiocache.Cache
    """Cache object to cache functions results."""
    cache_ttl: int
    """Time to live for the cache."""
    
    
    _built = False
    
    _BASE_API_URL = "https://api.hentaihaven.app/v1/"
    
    _logger: logging.Logger = logging.getLogger(__name__)
    
    _default_headers: typing.Mapping[str, str] = {
        "content-type": "application/x-www-form-urlencoded; charset=utf-8",
        "user-agent": "HH_xxx_APP",
        "warden": ""
    }
    
    _default_warden_body: typing.Mapping[str, typing.Any] = {
        "sdkInt": 33,
        "board": "goldfish_x86_64",
        "brand": "google",
        "display": "sdk_gphone_x86_64-userdebug 13 TE1A.220922.028 10190541 dev-keys",
        "fingerprint": "google/sdk_gphone_x86_64/emu64xa:13/TE1A.220922.028/10190541:userdebug/dev-keys",
        "manufacturer": "Google",
        "model": "sdk_gphone_x86_64"
    }
    
    
    def __init__(
        self, 
        token: str = None,
        *,
        cache: aiocache.Cache = None,
        cache_ttl: int = 1800,
        debug: bool = False,
        warden_body: typing.Mapping[str, typing.Any] = _default_warden_body
    ) -> None:
        """
        Initialize Hentai Heaven Client with the given parameters.

        Args:
            token (str, optional): Cloudflare token used to access Hentai Haven API.
            cache (aiocache.Cache, optional): Cache object to cache functions results.
            cache_ttl (int, optional): Time to live for the cache.
            debug (bool, optional): Whether the debug logs are being shown in stdout.
            warden_body (typing.Mapping[str, typing.Any], optional): Custom warden body with your device info.
            
        Note that in order to completely initialize client you 
        need to call `client.build()` function.
        """
        self.cache = cache
        self.debug = debug
        self.cache_ttl = cache_ttl
        self._default_warden_body = warden_body
        
        if token:
            self.token = token
            
            
    @property
    def token(self) -> str:
        """Cloudflare token used to access Hentai Haven API."""
        return self._default_headers["warden"]

    @token.setter
    def token(self, token: str) -> None:
        self._default_headers["warden"] = token
        
    
    @property
    def debug(self) -> bool:
        """Whether the debug logs are being shown in stdout."""
        return logging.getLogger("hhaven").level == logging.DEBUG
        
    
    @debug.setter
    def debug(self, debug: bool) -> None:
        logging.basicConfig()
        level = logging.DEBUG if debug else logging.NOTSET
        logging.getLogger("hhaven").setLevel(level)
    
    
    async def build(
        self, 
        token: str = None,
        *,
        validate_token: bool = True
    ) -> "Client":
        """
        Use this function to build a client.
        
        Args:
            token (str, optional): Token used to access Hentai Heaven API.
            validate_token (bool, optional): Validate token with test request to the API.
        Returns:
            Built client.
        """
        if token:
            self.token = token
            self._default_headers["warden"] = token
        
        if not self.token:
            await self.get_new_token(True)
        
        if validate_token:
            # Validate token
            await self._request("GET", "hentai/home", disable_logging = True)
        
        self._built = True
            
        return self
    

    async def _request(
        self,
        method: typing.Literal["GET", "POST"],
        path: str,
        headers: typing.Mapping[str, str] = _default_headers,
        data: typing.Mapping[str, typing.Any] = {},
        disable_logging: bool = False
    ) -> typing.Mapping[str, typing.Any]:
        async with aiohttp.ClientSession(headers = headers) as session:
            async with session.request(
                method = method,
                url = self._BASE_API_URL + path,
                data = data
            ) as r:
                response = await r.json()
                status = utility._get_status_from_response(response) or r.status
                
                if not disable_logging:
                    self._logger.debug("%s %s\n%s\n%s", method, r.url, json.dumps(data, separators=(",", ":")), response)
                
                if not str(status).startswith("2"):
                    return utility._raise_for_status(status)
                
                return response
        
    
    @requires_build
    @requires_token
    @cached
    async def home(self):
        """
        Get home page info.
        
        Returns:
            `models.HomePage` - [Docs](https://github.com)
        """
        data = await self._request("GET", "hentai/home")
        
        return models.HomePage(client = self, **data["data"])
    
    
    @requires_build
    @requires_token
    @cached
    async def search(self, query: str) -> list[models.PartialHentai]:
        """
        Search for hentai.
        
        Args:
            query (str): Query to search for.
        Returns:
            `list[models.PartialHentai]` - [Docs](https://github.com)
        """
        data = await self._request("GET", f"search?q={query}")
        
        # Nothing found
        if type(data["data"]) == str:
            return []
        
        return [models.PartialHentai(client = self, **post) for post in data["data"]]
    
    
    @requires_build
    @requires_token
    @cached
    async def get_hentai(self, id: int) -> models.Hentai:
        """
        Get full hentai info using it's ID. Raises `exceptions.HentaiEpisodeNotFound` if hentai was not found.
        
        Args:
            id (int): ID of the hentai.
        Returns:
            `models.Hentai` - [Docs](https://github.com)
        """
        data = await self._request("GET", f"hentai/{id}")
        
        # Not found
        if type(data["data"]) == str:
            raise exceptions.HentaiNotFound()
        
        return models.Hentai(client = self, **data["data"])
    
    
    @requires_build
    @requires_token
    @cached
    async def get_episode(self, id: int, hentai_id: int) -> models.HentaiEpisode:
        """
        Get full hentai episode info. Raises `exceptions.HentaiEpisodeNotFound` if episode was not found.
        
        Args:
            id (int): ID of the episode.
            hentai_id (int): ID of the hentai this episode belongs to.
        Returns:
            `models.HentaiEpisode` - [Docs](https://github.com)
        """
        data = await self._request("GET", f"hentai/{hentai_id}/episode/{id}")
        
        # Not found
        if type(data["data"]) == str:
            raise exceptions.HentaiEpisodeNotFound()
        
        return models.HentaiEpisode(**data["data"])
    
    
    @requires_build
    @requires_token
    @cached
    async def get_all_genres(self) -> models.HentaiGenre:
        """
        Get list of all hentai genres.
        
        Returns:
            `list[models.HentaiGenre]` - [Docs](https://github.com)
        """
        data = await self._request("GET", f"genre/all")
        
        return [models.HentaiGenre(client = self, **genre) for genre in data["data"]]

    
    @requires_build
    @requires_token
    @cached
    async def get_genre_page(self, id: int, page: int = 1) -> models.GenrePage:
        """
        Get page with list of hentai of the requested genre. Raises `exceptions.GenrePageNotFound` if page was not found.
        
        Args:
            id (int): ID of the genre you want to get page for.
            page (int, optional): Index of the page you want to get.
        Returns:
            `models.GenrePage` - [Docs](https://github.com)
        """
        data = await self._request("GET", f"genre/{id}?p={page}")
        
        # Not found
        if type(data["data"]) == str:
            raise exceptions.GenrePageNotFound()
        
        return models.GenrePage(client = self, **data["data"])
    
    
    @requires_build
    @requires_token
    @cached
    async def get_all_hentai(self, page: int = 1) -> models.HentaiPage:
        """
        Get page of all available hentai on the website. Raises `exceptions.HentaiPageNotFound` if page was not found.
        
        Args:
            page (int, optional): Index of the page you want to get.
        Returns:
            `models.HentaiPage` - [Docs](https://github.com)
        """
        data = await self._request("GET", f"hentai/all?p={page}")
        
        # Not found
        if type(data["data"]) == str:
            raise exceptions.HentaiPageNotFound()
        
        return models.HentaiPage(client = self, **data["data"])
    
    
    async def get_new_token(
        self, 
        apply: bool = True,
        *,
        body: typing.Mapping[str, typing.Any] = _default_warden_body,
        headers: typing.Mapping[str, str] = _default_headers
    ) -> str:
        """
        Obtain new Cloudflare token. 
        
        Args:
            apply (bool, optional): Apply received token to this client.
        
        Returns:
            Obtained token.
        """
        response = await self._request("POST", "warden", headers, body)
        token = response["data"]["token"]
        
        if apply: 
            self.token = token
            self._default_headers["warden"] = token
            
        return token



"""All HHaven models."""

from datetime import datetime
import pydantic, typing


__all__ = [
    "HentaiRating",
    "HentaiGenre",
    "PartialHentaiGenre",
    "GenrePage",
    "HentaiPage",
    "HentaiTag",
    "HentaiAuthor",
    "HentaiRelease",
    "HentaiEpisode",
    "PartialHentaiEpisode",
    "HomePage",
    "Hentai",
    "PartialHentai"
]


class HentaiRating(pydantic.BaseModel):
    """Information about hentai rating."""
    rating: float
    """Hentai rating."""
    votes: int
    """Number of people who voted when calculating the rating."""


class HentaiGenre(pydantic.BaseModel):
    """HentaiGenre object."""
    id: int
    """Genre ID."""
    name: str
    """
    Well formatted genre name.
    
    Examples: Yuri, Uncensored.
    """
    slug: str
    """
    Raw genre name.
    
    Examples: hentai-school, yuri.
    """
    count: int
    """Number of hentai of this genre."""
    thumbnail: typing.Optional[str] = None
    """
    Genre thumbnail url. Note that thumbnail is not present in `GenrePage.genre` object.
    
    Example: https://example.com/Yuri.jpg
    """
    client: typing.Any
    
    def __init__(self, **data):
        processed = {
            k.lower().replace("term_", ""): v
            for k, v in data.items()
        }
        super().__init__(**processed)
        
    async def page(self, page: int = 1) -> "GenrePage":
        """
        Get page with hentai of this genre.
        
        Args:
            page (int, optional): Index of the page you want to get.
        Returns:
            `models.GenrePage` - [Docs](https://github.com)
        """
        return await self.client.get_genre_page(self.id, page)


class PartialHentaiGenre(pydantic.BaseModel):
    """Short HentaiGenre object."""
    id: int
    """Genre ID."""
    name: str
    """
    Genre name.
    
    Examples: Yuri, Uncensored.
    """
    client: typing.Any
    
    def __init__(self, **data):
        processed = {
            k.lower().replace("term_", ""): v
            for k, v in data.items()
        }
        super().__init__(**processed)
        
    async def full(self) -> HentaiGenre:
        """
        Get full genre information.
        
        Returns:
            `models.HentaiGenre` - [Docs](https://github.com)
        """
        all_genres = await self.client.get_all_genres()
        return [genre for genre in all_genres if genre.id == self.id]
    
    async def page(self, page: int = 1) -> "GenrePage":
        """
        Get page with hentai of this genre.
        
        Args:
            page (int, optional): Index of the page you want to get.
        Returns:
            `models.GenrePage` - [Docs](https://github.com)
        """
        return await self.client.get_genre_page(self.id, page)
    
    
class GenrePage(pydantic.BaseModel):
    """Page with hentai of certain genre."""
    genre: HentaiGenre
    """Genre of the page."""
    hentai: list["PartialHentai"]
    """List of hentai on this page."""
    total_results: int
    """Total number of hentai on all pages of the genre."""
    index: int
    """Index of current page."""
    total_pages: int
    """Total pages available for this genre."""
    client: typing.Any
    
    def __init__(self, **data):
        data["genre"] = data.pop("term")
        data["hentai"] = data.pop("hentais")
        data["index"] = data.pop("current_page")
        
        for hentai in data["hentai"]:
            hentai["client"] = data.get("client")
        
        data["genre"]["client"] = data.get("client")
        
        super().__init__(**data)
        
    async def next(self) -> "GenrePage":
        """
        Get next page.
        
        Returns:
            `models.GenrePage` - [Docs](https://github.com)
        """
        return await self.client.get_genre_page(self.genre.id, self.index + 1)
    
    async def prev(self) -> "GenrePage":
        """
        Get previous page.
        
        Returns:
            `models.GenrePage` - [Docs](https://github.com)
        """
        return await self.client.get_genre_page(self.genre.id, self.index - 1)
        
        
class HentaiPage(pydantic.BaseModel):
    """Page with all hentai of the website."""
    hentai: list["PartialHentai"]
    """List of hentai on this page."""
    total_results: int
    """Total number of hentai on all pages."""
    index: int
    """Index of current page."""
    total_pages: int
    """Total pages available."""
    client: typing.Any
    
    def __init__(self, **data):
        data["hentai"] = data.pop("hentais")
        data["index"] = data.pop("current_page")
        
        for hentai in data["hentai"]:
            hentai["client"] = data.get("client")
        
        super().__init__(**data)
        
    async def next(self) -> "HentaiPage":
        """
        Get next page.
        
        Returns:
            `models.HentaiPage` - [Docs](https://github.com)
        """
        return await self.client.get_all_hentai(self.index + 1)
    
    async def prev(self) -> "HentaiPage":
        """
        Get previous page.
        
        Returns:
            `models.HentaiPage` - [Docs](https://github.com)
        """
        return await self.client.get_all_hentai(self.index - 1)
     
        
class HentaiTag(pydantic.BaseModel):
    """Hentai tag."""
    id: int
    """Tag ID."""
    name: str
    """
    Tag name.
    
    Examples: HD, Young
    """
    
    def __init__(self, **data):
        processed = {
            k.lower().replace("term_", ""): v
            for k, v in data.items()
        }
        super().__init__(**processed)
        
        
class HentaiAuthor(pydantic.BaseModel):
    """Hentai author object."""
    id: int
    """Author ID."""
    name: str
    """Author's name."""
    
    def __init__(self, **data):
        processed = {
            k.lower().replace("term_", ""): v
            for k, v in data.items()
        }
        super().__init__(**processed)
        
        
class HentaiRelease(pydantic.BaseModel):
    """Hentai release object."""
    id: int
    """Release ID."""
    name: str
    """
    Release name. In most cases it is the year when the hentai was released.
    
    Example: 2023
    """
    
    def __init__(self, **data):
        processed = {
            k.lower().replace("term_", ""): v
            for k, v in data.items()
        }
        super().__init__(**processed)
        
        
class HentaiEpisode(pydantic.BaseModel):
    """Hentai episode."""
    id: int
    """Episode ID."""
    name: str
    """
    Well formatted episode title.
    
    Example: Episode 1
    """
    slug: str
    """
    Raw title.
    
    Example: episode-1
    """
    date: datetime
    """Datetime object representing episode upload date."""
    content: str
    """
    Url to m3u8 file of the episode. Usually contains multiple quality levels:
    - 480p
    - 720p
    - 1080p
    
    Example: https://example.com/episode-name.m3u8
    """
    thumbnail: str
    """
    Episode thumbnail url.
    
    Example: https://example.com/episode_thumbnail.jpg
    """
    hentai_id: int
    """
    ID of the hentai this episode belongs to.
    """
    hentai_name: str
    """
    Raw name of the hentai this episode belongs to.
    
    Example: tsuma-shibori
    """
    hentai_title: str
    """
    Well formatted title of the hentai this episode belongs to.
    
    Example: Tsuma Shibori
    """
    hentai_views: int
    """Number of hentai views."""
    hentai_thumbnail: str
    """
    Hentai thumbnail url.
    
    Example: https://example.com/image.jpg
    """
    hentai_description: str
    """Full hentai description in English."""
    hentai_date: datetime
    """Datetime object representing hentai upload date."""
    hentai_rating: HentaiRating
    """Hentai rating object. Has `rating` and `votes` attributes."""
    hentai_tags: list[HentaiTag]
    """List of hentai tags."""
    hentai_title_alternative: str
    """
    Well formatted alternative title of the hentai this episode belongs to.
    
    Example: Nightmare×Deathscytheー前編...
    """
    hentai_genres: list[PartialHentaiGenre]
    """List of hentai genres."""
    hentai_authors: list[HentaiAuthor]
    """List of hentai authors."""
    hentai_releases: list[HentaiRelease]
    """List of hentai releases."""
    next_episode: typing.Optional["HentaiEpisode"] = None
    """Next hentai episode."""
    prev_episode: typing.Optional["HentaiEpisode"] = None
    """Previous hentai episode."""
    
    def __init__(self, **data):
        processed = {
            k.lower().replace("post_", "hentai_").replace("chapter_", ""): v
            for k, v in data.items()
        }
        processed["hentai_description"] = processed.pop("hentai_content")
        processed["date"] = datetime.strptime(processed["date"], "%Y-%m-%d %H:%M:%S")
        processed["hentai_date"] = datetime.strptime(processed["hentai_date"], "%Y-%m-%d %H:%M:%S")
        
        if data["next_episode"] is not None:
            for key, value in data.items():
                if key not in ("id", "name", "thumbnail", "date", "slug"):
                    if key == "next_episode" or key == "prev_episode":
                        data["next_episode"][key] = None
                    else:
                        data["next_episode"][key] = value
                    
        if data["prev_episode"] is not None:
            for key, value in data.items():
                if key not in ("id", "name", "thumbnail", "date", "slug"):
                    if key == "next_episode" or key == "prev_episode":
                        data["prev_episode"][key] = None
                    else:
                        data["prev_episode"][key] = value
                        
        for genre in processed["hentai_genres"]:
            genre["client"] = data.get("client")
        
        super().__init__(**processed)
        
        
class PartialHentaiEpisode(pydantic.BaseModel):
    """Simplified HentaiEpisode object."""
    id: int
    """Episode ID."""
    name: str
    """
    Well formatted episode title.
    
    Example: Episode 1
    """
    slug: str
    """
    Raw title.
    
    Example: episode-1
    """
    date: datetime
    """Datetime object representing episode upload date."""
    thumbnail: str
    """
    Episode thumbnail url.
    
    Example: https://example.com/episode_thumbnail.jpg
    """
    hentai_id: int
    """
    ID of the hentai this episode belongs to.
    """
    hentai_name: str
    """
    Raw name of the hentai this episode belongs to.
    
    Example: tsuma-shibori
    """
    hentai_title: str
    """
    Well formatted title of the hentai this episode belongs to.
    
    Example: Tsuma Shibori
    """
    hentai_thumbnail: str
    """
    Hentai thumbnail url.
    
    Example: https://example.com/image.jpg
    """
    hentai_description: str
    """Full hentai description in English."""
    client: typing.Any
    
    def __init__(self, **data):
        processed = {
            k.lower().replace("post_", "hentai_").replace("chapter_", ""): v
            for k, v in data.items()
        }
        processed["hentai_description"] = processed.pop("hentai_content")
        processed["date"] = datetime.strptime(processed["date"], "%Y-%m-%d %H:%M:%S")
        
        super().__init__(**processed)
        
    async def full(self) -> HentaiEpisode:
        """
        Get full episode object.
        
        Returns:
            `models.HentaiEpisode` - [Docs](https://github.com)
        """
        return await self.client.get_episode(self.id, self.hentai_id)

        
class Hentai(pydantic.BaseModel):
    """Hentai object."""
    id: int
    """ID of the hentai."""
    name: str
    """
    Raw hentai name.
    
    Example: tsuma-shibori
    """
    title: str
    """
    Well formatted title.
    
    Example: Tsuma Shibori
    """
    views: int
    """Number of views."""
    thumbnail: str
    """
    Hentai thumbnail url.
    
    Example: https://example.com/image.jpg
    """
    date: datetime
    """Datetime object representing hentai upload date."""
    description: str
    """Full description in English."""
    rating: HentaiRating
    """Hentai rating object. Has `rating` and `votes` attributes."""
    tags: list[HentaiTag]
    """List of hentai tags."""
    title_alternative: str
    """
    Well formatted alternative title.
    
    Example: Nightmare×Deathscytheー前編...
    """
    genres: list[PartialHentaiGenre]
    """List of hentai genres."""
    authors: list[HentaiAuthor]
    """List of hentai authors."""
    releases: list[HentaiRelease]
    """List of releases."""
    episodes: list[PartialHentaiEpisode]
    """List of episodes."""
    client: typing.Any
    
    def __init__(self, **data):
        processed = {
            k.lower().replace("post_", ""): v
            for k, v in data.items()
        }
        processed["description"] = processed.pop("content")
        processed["date"] = datetime.strptime(processed["date"], "%Y-%m-%d %H:%M:%S")
        
        for episode in processed["episodes"]:
            episode["client"] = data.get("client")
            episode["hentai_id"] = processed.get("id")
            episode["hentai_name"] = processed.get("name")
            episode["hentai_title"] = processed.get("title")
            episode["hentai_thumbnail"] = processed.get("thumbnail")
            episode["hentai_content"] = processed.get("description")
                    
        for genre in processed["genres"]:
            genre["client"] = data.get("client")
        
        super().__init__(**processed)


class PartialHentai(pydantic.BaseModel):
    """Simplified Hentai object."""
    id: int
    """ID of the hentai."""
    name: str
    """
    Raw hentai name.
    
    Example: tsuma-shibori
    """
    title: str
    """
    Well formatted hentai title.
    
    Example: Tsuma Shibori
    """
    thumbnail: str
    """
    Hentai thumbnail url.
    
    Example: https://example.com/image.jpg
    """
    client: typing.Any
    
    def __init__(self, **data):
        processed = {
            k.lower().replace("post_", ""): v
            for k, v in data.items()
        }
        super().__init__(**processed)
        
    async def full(self) -> Hentai:
        """
        Get full hentai object.
        
        Returns:
            `models.Hentai` - [Docs](https://github.com)
        """
        return await self.client.get_hentai(self.id)


class HomePage(pydantic.BaseModel):
    """Home page information."""
    
    client: typing.Any
    last: list[PartialHentai]
    """Last released hentai."""
    yuri: list[PartialHentai]
    """List of popular yuri hentai."""
    ecchi: list[PartialHentai]
    """List of popular ecchi hentai."""
    incest: list[PartialHentai]
    """List of popular incest hentai."""
    tentacle: list[PartialHentai]
    """List of popular tentacle hentai."""
    uncensored: list[PartialHentai]
    """List of popular uncensored hentai."""
    trending_month: list[PartialHentai]
    """List of trending month hentai."""
    last_episodes: list[PartialHentaiEpisode]
    """Last released hentai episodes."""
    
    def __init__(self, **data):
        for key, value in data.items():
            if isinstance(value, list):
                for item in value:
                    item["client"] = data.get("client")
                    
        super().__init__(**data)
        


"""All HHaven exceptions."""

import typing


__all__ = [
    "HHavenException",
    "HHavenRateLimited",
    "HHavenNotFound",
    "TokenException",
    "HentaiNotFound",
    "GenrePageNotFound",
    "HentaiPageNotFound",
    "HentaiEpisodeNotFound"
]


class HHavenException(Exception):
    """Base Hentai Haven exception."""
    
    status: int = 0
    message: str = ""
    
    def __init__(
        self,
        message: str = None,
        status: int = status
    ) -> None:
        self.status = status
        self.message = message or self.message
        
        if self.status:
            self.message = f"[{self.status}] {self.message}"
        
        super().__init__(self.message)
    
    @property
    def response(self) -> typing.Mapping[str, typing.Any]:
        return {"status": self.status, "message": self.message, "data": None}
    
    
class HHavenRateLimited(HHavenException):
    """You were rate limited. Please try again later."""
    
    status = 429
    message = "You were rate limited. Please try again later."
    
    
class HHavenNotFound(HHavenException):
    """Requested resource does not exist."""
    
    status = 404
    message = "Requested resource does not exist."
    

class TokenException(HHavenException):
    """Invalid Hentai Haven token."""
    
    status = 502
    message = "Invalid Hentai Haven token."
    
    
class HentaiNotFound(HHavenNotFound):
    """Unable to find hentai with the ID you provided."""
    
    status = 404
    message = "Unable to find hentai with the ID you provided."
    
    
class GenrePageNotFound(HHavenNotFound):
    """Unable to find genre page with the ID and index you provided."""
    
    status = 404
    message = "Unable to find genre page with the ID and index you provided."
    

class HentaiPageNotFound(HHavenNotFound):
    """Unable to find hentai page with this index."""
    
    status = 404
    message = "Unable to find hentai page with this index."
    
    
class HentaiEpisodeNotFound(HHavenNotFound):
    """Unable to find hentai episode with the IDs you provided."""
    
    status = 404
    message = "Unable to find hentai episode with the IDs you provided."



"""Library decorators."""

from . import exceptions


__all__ = []


def requires_build(func):
    def wrapper(self, *args, **kwargs):
        if not self._built:
            raise exceptions.HHavenException("Hentai Haven client was not fully initialized. Remember to use client.build() to complete client initialization.", 0)
        return func(self, *args, **kwargs)
    return wrapper


def requires_token(func):
    def wrapper(self, *args, **kwargs):
        if not self.token:
            raise exceptions.TokenException("Hentai Haven token is missing. Use client.get_new_token() to obtain it.", 0)
        return func(self, *args, **kwargs)
    return wrapper


def cached(func):
    """Cache async function."""
    async def wrapper(self, *args, **kwargs):
        if self.cache:
            key = f"{func.__name__}_{args}_{kwargs}"
            cached = await self.cache.get(key)
            if cached is not None:
                return cached

        result = await func(self, *args, **kwargs)

        if self.cache:
            await self.cache.set(key, result, ttl=self.cache_ttl)

        return result

    return wrapper



"""Some useful utility."""

import typing

from . import exceptions


__all__ = []


def _raise_for_status(status: int):
    """Raise exception for status."""
    
    if status == 502:
        raise exceptions.TokenException()
    elif status == 429:
        raise exceptions.HHavenRateLimited()
    elif status == 404:
        raise exceptions.HHavenNotFound()
    else:
        raise exceptions.HHavenException("Something went wrong.", status)
    

def _get_status_from_response(response: typing.Mapping[str, typing.Any]) -> int | None:
    """Determine status using response text."""
    
    data = response.get("data", None)
    if data and type(data) == str:
        if data[:3].isdigit(): 
            return int(data[:3])
        
    return None



"""Run setuptools."""
from setuptools import find_packages, setup

setup(
    name = "hhaven",
    version = "0.1.4",
    author = "jokelbaf",
    author_email = "jokelbaf@gmail.com",
    description = "A well-documented and typed API wrapper for Hentai Haven.",
    keywords = "hentai-haven hentai".split(),
    url = "https://github.com/jokelbaf/hhaven",
    project_urls = {
        "Documentation": "https://jokelbaf.github.io/hhaven",
        "API Reference": "https://jokelbaf.github.io/hhaven-api-reference",
        "Issue tracker": "https://github.com/jokelbaf/hhaven/issues",
    },
    packages = find_packages(exclude=["tests.*"]),
    python_requires = ">=3.8",
    install_requires = ["aiohttp", "pydantic", "aiocache"],
    include_package_data = True,
    package_data = {"hhaven": ["py.typed"]},
    long_description = open("README.md", encoding="utf-8").read(),
    long_description_content_type = "text/markdown",
    license = "MIT",
    classifiers = [
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
        "Typing :: Typed",
    ],
)